#display.py
#Copyright (C) 2006 Evil Mr Henry
#This file is part of The Monster's Tower.

#The Monster's Tower is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#The Monster's Tower is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with The Monster's Tower; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#This file handles all the opengl stuff.

from OpenGL import GL
from OpenGL import GLU
import pygame
from os import listdir

import level
import math

#Honestly, I just pulled this off one of the nehe tutorials. I think some of
#it is unneeded.
def setup_opengl((width, height)):
    GL.glEnable(GL.GL_TEXTURE_2D)
    GL.glDisable(GL.GL_LIGHTING)
    if height==0:
        height=1.0
    GL.glViewport(0, 0, width, height)
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glDepthFunc(GL.GL_LESS)
    GL.glEnable(GL.GL_DEPTH_TEST)
    GL.glLoadIdentity()
    GLU.gluPerspective(45, float(width)/height, 0.1, 100.0)
    GL.glMatrixMode(GL.GL_MODELVIEW)
    GL.glLoadIdentity()


floor_squares = []
wall_n_squares = []
wall_s_squares = []
wall_e_squares = []
wall_w_squares = []
ceiling_squares = []

def draw_tile(sq, height):
    GL.glBegin(GL.GL_POLYGON)
    GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(sq[0], -sq[1], height)
    GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(sq[0]+1, -sq[1], height)
    GL.glTexCoord2f(1.0, -1.0); GL.glVertex3f(sq[0]+1, -sq[1]-1, height)
    GL.glTexCoord2f(0.0, -1.0); GL.glVertex3f(sq[0], -sq[1]-1, height)
    GL.glEnd()

def draw_rot_tile(loc, size, direction, height):
    upperleft = direction + 3.0*math.pi/4.0
    lowerleft = direction + 5.0*math.pi/4.0
    lowerright = direction + 7.0*math.pi/4.0
    upperright = direction + 9.0*math.pi/4.0

    upperleft = (math.cos(upperleft)*size*math.sqrt(2)/2.0, math.sin(upperleft)*size*math.sqrt(2)/2.0)
    lowerleft = (math.cos(lowerleft)*size*math.sqrt(2)/2.0, math.sin(lowerleft)*size*math.sqrt(2)/2.0)
    lowerright = (math.cos(lowerright)*size*math.sqrt(2)/2.0, math.sin(lowerright)*size*math.sqrt(2)/2.0)
    upperright = (math.cos(upperright)*size*math.sqrt(2)/2.0, math.sin(upperright)*size*math.sqrt(2)/2.0)
    GL.glBegin(GL.GL_POLYGON)
    #print upperleft, lowerleft, lowerright, upperright
    GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(loc[0]/10.0+upperleft[0], -loc[1]/10.0+upperleft[1], height)
    GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(loc[0]/10.0+lowerleft[0], -loc[1]/10.0+lowerleft[1], height)
    GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(loc[0]/10.0+lowerright[0], -loc[1]/10.0+lowerright[1], height)
    GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(loc[0]/10.0+upperright[0], -loc[1]/10.0+upperright[1], height)
    GL.glEnd()

global polycount

def draw_frame(loc, zoom, lookdir, cur_level, health, flash, pldir):
    global polycount
    polycount = 0
    GL.glLoadIdentity()
    GL.glTranslatef(-loc[0]/10, loc[1]/10, -zoom)
    GL.glClear( GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT )

    #draw floor squares
    GL.glBindTexture(GL.GL_TEXTURE_2D, 3)
    for sq in floor_squares:
        polycount+=1
        draw_tile(sq, 1.0)

    #draw ceiling squares
    GL.glBindTexture(GL.GL_TEXTURE_2D, 1)
    for sq in ceiling_squares:
        polycount+=1
        draw_tile(sq, 4.0)

    GL.glBindTexture(GL.GL_TEXTURE_2D, 2)
    for sq in wall_n_squares:
        if sq[1]*10 < loc[1]: continue
        for i in range(4):
            polycount+=1
            GL.glBegin(GL.GL_POLYGON)
            GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(sq[0], -sq[1], i)
            GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(sq[0]+1, -sq[1], i)
            GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(sq[0]+1, -sq[1], i+1)
            GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(sq[0], -sq[1], i+1)
            GL.glEnd()
    for sq in wall_s_squares:
        if sq[1]*10 > loc[1]: continue
        for i in range(4):
            polycount+=1
            GL.glBegin(GL.GL_POLYGON)
            GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(sq[0], -sq[1]-1, i)
            GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(sq[0]+1, -sq[1]-1, i)
            GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(sq[0]+1, -sq[1]-1, i+1)
            GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(sq[0], -sq[1]-1, i+1)
            GL.glEnd()
    for sq in wall_e_squares:
        if sq[0]*10 > loc[0]: continue
        for i in range(4):
            polycount+=1
            GL.glBegin(GL.GL_POLYGON)
            GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(sq[0]+1, -sq[1], i)
            GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(sq[0]+1, -sq[1]-1, i)
            GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(sq[0]+1, -sq[1]-1, i+1)
            GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(sq[0]+1, -sq[1], i+1)
            GL.glEnd()
    for sq in wall_w_squares:
        if sq[0]*10 < loc[0]: continue
        for i in range(4):
            polycount+=1
            GL.glBegin(GL.GL_POLYGON)
            GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(sq[0], -sq[1], i)
            GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(sq[0], -sq[1]-1, i)
            GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(sq[0], -sq[1]-1, i+1)
            GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(sq[0], -sq[1], i+1)
            GL.glEnd()



    #flashlights
    GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
    clipping_qual = 20
    num_beams = 20
    beam_width = level.maps[cur_level].enemy_list[0].sightwidth*(6.283185)/(20.0*360.0)
    for enemy in level.maps[cur_level].enemy_list:
        for i in range(-num_beams, num_beams):
            ray_length = enemy.sightrange /10.0
            for j in range(0, clipping_qual):
                if (level.maps[cur_level].grid[int(enemy.loc[1]/10.0
                -math.sin(enemy.sightdir+i*beam_width
                -beam_width/2.0)*ray_length*j/clipping_qual)]
                [int(enemy.loc[0]/10.0+math.cos(enemy.sightdir+
                i*beam_width-beam_width/2.0)*ray_length*j/clipping_qual)]
                == "0"):
                    ray_length = (ray_length * (j+1) * 1.0) / clipping_qual
                    break
            polycount+=1
            draw_ray(enemy.loc, enemy.sightdir+i*beam_width-beam_width/2.0,
                    beam_width, ray_length)


    #healthbox
    #Yes, it's really that complicated...
    GL.glDisable(GL.GL_DEPTH_TEST)
    GL.glDisable(GL.GL_TEXTURE_2D)
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glPushMatrix()
    GL.glLoadIdentity()
    GL.glOrtho(0,640,0,480,-1,1)
    GL.glMatrixMode(GL.GL_MODELVIEW)
    GL.glPushMatrix()
    GL.glLoadIdentity()
    GL.glTranslated(5,5,0)
    GL.glColor3f (0.9, 0.1, 0.1)
    GL.glLineWidth (1)

    GL.glBegin (GL.GL_LINE_LOOP)
    GL.glVertex2i(0, 0)
    GL.glVertex2i (20, 0)
    GL.glVertex2i (20, 100)
    GL.glVertex2i (0, 100)
    GL.glEnd()
    if flash:
        GL.glColor3f (0.8, 0.8, 0.8)
    else:
        GL.glColor3f (0.8, 0.1, 0.1)
    GL.glBegin(GL.GL_POLYGON)
    GL.glVertex2i(0, 0)
    GL.glVertex2i (20, 0)
    GL.glVertex2i (20, health/2)
    GL.glVertex2i (0, health/2)
    GL.glEnd()

    GL.glColor3f (1.0, 1.0, 1.0)
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glPopMatrix()
    GL.glMatrixMode(GL.GL_MODELVIEW)
    GL.glPopMatrix()
    GL.glEnable(GL.GL_DEPTH_TEST)
    GL.glEnable(GL.GL_TEXTURE_2D)

    #blood
    GL.glBindTexture(GL.GL_TEXTURE_2D, 4)
    GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE)
    GL.glEnable(GL.GL_BLEND);
    i = 1.001
    for blood in level.enemy.blood_array:
        polycount+=1
        GL.glBegin(GL.GL_POLYGON)
        GL.glTexCoord2f(0.0, 0.0); GL.glVertex3f(blood.loc[0]/10-0.3, -blood.loc[1]/10-0.3, i)
        GL.glTexCoord2f(1.0, 0.0); GL.glVertex3f(blood.loc[0]/10+0.3, -blood.loc[1]/10-0.3, i)
        GL.glTexCoord2f(1.0, 1.0); GL.glVertex3f(blood.loc[0]/10+0.3, -blood.loc[1]/10+0.3, i)
        GL.glTexCoord2f(0.0, 1.0); GL.glVertex3f(blood.loc[0]/10-0.3, -blood.loc[1]/10+0.3, i)
        GL.glEnd()
        i += 0.001

    #enemies
    GL.glBlendFunc(GL.GL_DST_COLOR,GL.GL_ZERO)
    GL.glDisable(GL.GL_DEPTH_TEST)
    for enemy in level.maps[cur_level].enemy_list:
        GL.glBlendFunc(GL.GL_DST_COLOR,GL.GL_ZERO)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 7)
        draw_rot_tile(enemy.loc, 0.5, enemy.sightdir, i)
        GL.glBlendFunc(GL.GL_ONE, GL.GL_ONE)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 5)
        draw_rot_tile(enemy.loc, 0.5, enemy.sightdir, i)
        polycount+=2
        i += 0.001
    #player
    GL.glBlendFunc(GL.GL_DST_COLOR,GL.GL_ZERO)
    GL.glBindTexture(GL.GL_TEXTURE_2D, 8)
    draw_rot_tile(loc, 0.5, pldir, i)
    GL.glBlendFunc(GL.GL_ONE, GL.GL_ONE)
    GL.glBindTexture(GL.GL_TEXTURE_2D, 6)
    draw_rot_tile(loc, 0.5, pldir, i)
    polycount+=2
    GL.glDisable(GL.GL_BLEND);
    GL.glEnable(GL.GL_DEPTH_TEST)

    return polycount

#start_pos should take player and monster locs with no conversion,
#ray_dir is in radians from straight up,
#ray width is also in radians.
#ray length is measured from the ray start along a side, and is in tiles.
def draw_ray(start_pos, ray_dir, ray_width, ray_length):
    ray_width = ray_width / 2.0
    GL.glBegin(GL.GL_POLYGON)

    GL.glTexCoord2f((start_pos[0]%10)/10.0, -(start_pos[1]%10)/10.0);
    GL.glVertex3f(start_pos[0]/10.0, -start_pos[1]/10.0, 1.001)

    GL.glTexCoord2f((start_pos[0]%10)/10.0 +
    math.cos(ray_dir-ray_width)*ray_length,
    -((start_pos[1]%10)/10.0-math.sin(ray_dir-ray_width)*ray_length));
    GL.glVertex3f(start_pos[0]/10.0+math.cos(ray_dir-ray_width)*ray_length,
    -start_pos[1]/10.0+math.sin(ray_dir-ray_width)*ray_length, 1.001)

    GL.glTexCoord2f((start_pos[0]%10)/10.0 +
    math.cos(ray_dir+ray_width)*ray_length,
    -((start_pos[1]%10)/10.0-math.sin(ray_dir+ray_width)*ray_length));
    GL.glVertex3f(start_pos[0]/10.0+math.cos(ray_dir+ray_width)*ray_length,
    -start_pos[1]/10.0+math.sin(ray_dir+ray_width)*ray_length, 1.001)
    GL.glEnd()


def load_level(index):
    global floor_squares; global wall_n_squares; global wall_s_squares
    global wall_e_squares; global wall_w_squares; global ceiling_squares
    floor_squares = []; wall_n_squares = []; wall_s_squares = []
    wall_e_squares = []; wall_w_squares = []; ceiling_squares = []

    load_pixs((level.maps[index].floor_pic, level.maps[index].ceil_pic,
        level.maps[index].wall_pic,
        level.maps[index].floor_pic[:-4]+"-dark.png",
        "blood.png", "human.png", "monster.png", "humanm.png", "monsterm.png"))



    for i in range(len(level.maps[index].grid)):
        for j in range(len(level.maps[index].grid[i])):
            if level.maps[index].grid[i][j] == '1':
                floor_squares.append((j, i))
            elif level.maps[index].grid[i][j] == '0':
                ceiling_squares.append((j, i))

                if i != 0:
                    if level.maps[index].grid[i-1][j] == '1':
                        wall_n_squares.append((j, i))

                if i != level.maps[index].size[1]-1:
                    if level.maps[index].grid[i+1][j] == '1':
                        wall_s_squares.append((j, i))

                if j != 0:
                    if level.maps[index].grid[i][j-1] == '1':
                        wall_w_squares.append((j, i))
                if j != level.maps[index].size[0]-1:
                    if level.maps[index].grid[i][j+1] == '1':
                        wall_e_squares.append((j, i))

pixs = []
pixs2 = []

#Some of this is from the 6th nehe tutorial. Seems a bit "unpythony" to me...
def load_pixs(pic_list):
    global pixs; global pixs2
    pixs = []; pixs2 = []

    for i in range(len(pic_list)):
        pixs2.append(pygame.image.load("../images/"+pic_list[i]).convert_alpha())
        pixs.append(pygame.image.tostring(pixs2[-1], "RGBX", 1))
        GL.glBindTexture(GL.GL_TEXTURE_2D, i)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, pixs2[i].get_width(),
        pixs2[i].get_height(), 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, pixs[i])
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER,
            GL.GL_NEAREST)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER,
            GL.GL_NEAREST_MIPMAP_LINEAR)
        GLU.gluBuild2DMipmaps(GL.GL_TEXTURE_2D, 3, pixs2[i].get_width(),
            pixs2[i].get_height(), GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, pixs[i])


